#ifndef _NV_FREEBSD_H_
#define _NV_FREEBSD_H_
#include "nv.h"
#include <machine/resource.h>
#include <sys/bus.h>
#include <nvtypes.h>
#include <sys/types.h>
#include <sys/malloc.h>
#include <sys/ioccom.h>
#include <sys/lock.h>
#ifdef __KERNEL__
#include <vm/vm.h>
#include <vm/pmap.h>
#include <vm/vm_map.h>
#include <vm/vm_object.h>
#include <vm/vm_page.h>
#include "inc.h"
#endif

struct ioc_tunnel {
	unsigned long address;
	unsigned long cmd;
};

#define NV_IOCTL_TUNNEL		_IOWR('G', 1, struct ioc_tunnel) 

#define DUMMY_IOCTL		_IO('G', 1)

/*
 * XXX DEPRECATED
 * struct ioc_key {
 *         unsigned int key;
 * };
 * #define NV_IOCTL_MMAP_KEY	_IOWR('G', 2, struct ioc_key)
 */

/* first 24 bits are for offset next 3 are for key */
#define KEY_MASK 0x7
/* XXX these next two have to match */
#define NUM_BITS 24
#define NV_ALLOC_OFFSET(x) (x)&0xFFFFFF
#define NV_ALLOC_FROM_KEY(x) (NV_MMAP_ALLOCATION_OFFSET|(((x)&KEY_MASK)<<NUM_BITS))
#define NV_KEY_FROM_ALLOC(x) (((x)>>NUM_BITS)&KEY_MASK)


#ifdef __KERNEL__

MALLOC_DECLARE(M_NV);

extern unsigned long		last_ioctl_command;

#ifndef DEBUG
#  define dprintf(fmt, args...)
#  define dvprintf(fmt, args...)
#else
#  ifdef DEBUG_VERBOSE
#    define dvprintf(fmt, args...) printf(fmt, ## args)
#  else
#    define dvprintf(fmt, args...)
#  endif
#  define dprintf(fmt, args...) printf(fmt, ## args)
#endif

struct nv_alloc {
	TAILQ_ENTRY(nv_alloc)	entries;
	unsigned long		vaddr; /* kernel virtual address */
	unsigned long		uvaddr; /* user virtual address */
	int			pages, used;
	unsigned long		*page_table;
};

TAILQ_HEAD(nv_alloc_head, nv_alloc);

#define MAX_KEYS 8

typedef struct {
	/* freebsd specific stuff */
	struct nv_alloc		*allocs[MAX_KEYS];
	struct nv_alloc		*last;

	struct {
		int rid;
		struct resource *res;
	} res[3];
#define reg res[0]
#define ffb res[1]
#define intr res[2]

	dev_t  dev;

	void *intr_cookie;
} nv_freebsd_state_t;

void	nv_release_resources (device_t dev);
int	nv_map_device (device_t device, int devnum);

#define nv_device_get_softc(device) ((nv_state_t *)device_get_softc(device))
#define nv_device_get_softc_os(device) (nv_freebsd_state_t *)(nv_device_get_softc(device)->os_state)

struct nv_alloc *nv_find_alloc __P((nv_freebsd_state_t *, unsigned long, unsigned long *, int));

int nv_alloc_get_key __P((nv_state_t *nv));

#endif

/*
int   nv_vtop(nv_ioctl_memory_vtop_t *params, int in_kernel);
void *nv_find_kernel_mapping(nv_state_t *, void *address);

void  nv_free_all_unused_clients(nv_state_t *nv, int pid, void *file);
*/

#endif /* _NV_FREEBSD_H */
