#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/errno.h>
#include <sys/conf.h>
#include <machine/bus.h>
#include <sys/bus.h>
#include <sys/rman.h>
#include <pci/pcireg.h>
#include <pci/pcivar.h>
#include <machine/stdarg.h>
#include <sys/lock.h>
#include <sys/proc.h>
#include <vm/vm_param.h>
#include <vm/vm.h>
#include <vm/pmap.h>
#include <vm/vm_map.h>
#include <machine/pmap.h>
#include <vm/vm_kern.h>
#include <vm/vm_extern.h>

#include "nv-freebsd.h"
#include "nvtypes.h"
#define RM_HEADERS_MINIMAL
#include "nvrm.h"
#include "nv_ref.h"
#include "nv_dev.h"

static int nv_probe		__P((device_t));
static int nv_attach		__P((device_t));
static int nv_detach		__P((device_t));
static int nv_shutdown		__P((device_t));

#define NV_PAGE_ALIGN(addr)    ( ((addr) + PAGE_SIZE - 1) / PAGE_SIZE)
#define NV_MASK_OFFSET(addr)  ( (addr) & (PAGE_SIZE - 1) )

device_t	nv_devices[NV_MAX_DEVICES];
nv_client_t	nv_clients[NV_MAX_CLIENTS];

/*
 * Hardware access macros for the kernel driver only
 * The resource manager part of the driver uses different values here
 */

/* HW register access macros */
#define REG_WR08(a,d)   nv->regs_u->Reg008[(a)]   = (U008) (d)
#define REG_WR16(a,d)   nv->regs_u->Reg016[(a)/2] = (U016) (d)
#define REG_WR32(a,d)   nv->regs_u->Reg032[(a)/4] = (U032) (d)

#define REG_RD08(a)     nv->regs_u->Reg008[(a)]
#define REG_RD16(a)     nv->regs_u->Reg016[(a)/2]
#define REG_RD32(a)     nv->regs_u->Reg032[(a)/4]

int nv_map_device (device, devnum)
	device_t		device;
	int			devnum;
{
	nv_state_t		*sc_old = nv_device_get_softc (device);
	nv_freebsd_state_t	*sc = sc_old->os_state;

	sc->reg.res = bus_alloc_resource (device, SYS_RES_MEMORY, &sc->reg.rid,
			0, ~0, 1, RF_ACTIVE);
	sc->ffb.res = bus_alloc_resource (device, SYS_RES_MEMORY, &sc->ffb.rid,
			0, ~0, 1, RF_ACTIVE);

	if (!(sc->ffb.res && sc->reg.res)) {
		printf ("nvidia%d: can't map I/O memory!\n", devnum);
		return (-1);
	}

	sc_old->regs = rman_get_virtual (sc->reg.res);
	printf ("NV%d: regs=%p\n", devnum, sc_old->regs);
	sc_old->regs_u = rman_get_virtual (sc->reg.res);
	sc_old->fb = rman_get_virtual (sc->ffb.res);
	printf ("NV%d: regs=%p\n", devnum, sc_old->fb);
	sc_old->fb_u = rman_get_virtual (sc->ffb.res);

	sc->intr.res = bus_alloc_resource (device, SYS_RES_IRQ, &sc->intr.rid,
			0, ~0, 1, RF_ACTIVE|RF_SHAREABLE);
	if (sc->intr.res == NULL) {
		printf ("nvidia%d: interrupt request failed\n", devnum);
		return (-1);
	}

	return (0);
}

void nv_set_hotkey_occurred_flag(void)
{
}

int
nv_int10h_call(
    nv_state_t *nv,
    U032 *eax,
    U032 *ebx,
    U032 *ecx,
    U032 *edx,
    VOID *buffer
)
{
    if ( (*eax == 0x4f14) && (*ebx == 0x0186) )
    {
        *eax = 0x004f;
        *ecx = ((REG_RD32(0x101000)>>16)&0xf);
    }
    return 0x0;
}

struct nv_alloc *nv_find_alloc (sc, address, kernaddr, align)
	nv_freebsd_state_t	*sc;
	unsigned long		address;
	unsigned long		*kernaddr;
	int			align;
{
	int			i;
	struct nv_alloc		*at;
	unsigned long		offset;

	if (kernaddr)
		*kernaddr = 0x0;

#define NV_PAGE_MASK (~(unsigned long)(PAGE_SIZE-1))

	if (align)
		address &= NV_PAGE_MASK;

	if (address >= VM_MIN_KERNEL_ADDRESS) {
		printf ("daymn\n");
		return (NULL);
	}

	for (i = 0; i < MAX_KEYS; ++i) {
		at = sc->allocs[i];
		if (!at)
			continue;
		if (address >= at->uvaddr && address < (at->uvaddr+ptoa(at->pages))) {
			offset = address - at->uvaddr;
			if (kernaddr)
				*kernaddr = at->vaddr + offset;
			return (0);
		}
	}
done:
	dprintf ("could not find map for vm 0x%lx\n", address);
	return (NULL);
}

void *nv_find_kernel_mapping (nv, address)
	nv_state_t		*nv;
	void			*address;
{
	nv_freebsd_state_t	*sc = nv->os_state;
	unsigned long		kernaddr;

	dprintf ("nv_find_kernel_mapping (%p)\n", address);
	if (inc_entry())
		return (0x0);

	nv_find_alloc (sc, (unsigned long)address, &kernaddr, 1);

	dprintf ("returned 0x%08lx\n", kernaddr);
	
	return ((void *)kernaddr);
}

int nv_agp_init (nv, phys_start, linear_start, agp_limit, config)
	nv_state_t		*nv;
	VOID			**phys_start;
	VOID			**linear_start;
	VOID			*agp_limit;
	U032			config;
{
	return -1;
}

int nv_agp_teardown (nv)
	nv_state_t		*nv;
{
	return -1;
}

int nv_agp_translate_address (nv, base, index, paddr)
	nv_state_t		*nv;
	void			*base;
	U032			index;
	U032			*paddr;
{
	return 0x1;
}

int nv_agp_retrieve_mapping (nv, address, agp_address)
	nv_state_t		*nv;
	void			*address;
	unsigned long		*agp_address;
{
	return 0x1;
}

void nv_lock_grab (pflags)
	unsigned long		*pflags;
{
}

void nv_lock_release (flags)
	unsigned long		flags;
{
}

int nv_alloc_get_key (nv)
	nv_state_t			*nv;
{
	nv_freebsd_state_t		*sc = nv->os_state;
	int				key;
	struct nv_alloc			*alloc;

	dprintf ("nv_alloc_get_key\n");

	for (key = 0; key < MAX_KEYS; ++key) {
		if (sc->allocs[key] == 0) {
			alloc = sc->last;
			if (!alloc)
				return (-1);
			sc->allocs[key] = alloc;
			return (key);
		}
	}

	return (-1);
}

int nv_alloc_pages (nv, pAddress, page_count, agp_memory, class, priv_data)
	nv_state_t			*nv;	
	void				**pAddress;
	unsigned int			page_count;
	unsigned int			agp_memory;
	unsigned int			class;
	void				**priv_data;
{
	nv_freebsd_state_t		*sc = nv->os_state;
	struct nv_alloc			*ptr = NULL;
	int				i;

	dprintf ("nv_alloc_pages %d\n", page_count);

	if (inc_entry())
		return (-1);

	if (!agp_memory) {
		ptr = malloc (sizeof(struct nv_alloc), M_NV, 0);
		if (ptr == NULL) {
			dprintf ("damn\n");
			return (-1);
		}
		ptr->page_table = malloc (sizeof(unsigned long)*(page_count+1), M_NV, 0);
		if (ptr->page_table == NULL) {
			dprintf ("damn\n");
			return (-1);
		}
		ptr->vaddr = kmem_alloc (kernel_map, ptoa(page_count));
		printf ("kmem_alloc: %08lx\n", ptr->vaddr);
		ptr->pages = page_count;
		ptr->used = 0;
	
		if (pAddress)
			*pAddress = (void *)ptr->vaddr;

		ptr->page_table[0] = page_count;
		ptr->page_table++;
		for (i = 0; i < page_count; ++i) {
			ptr->page_table[i] = vtophys (ptr->vaddr + ptoa(i));
		}

		*pAddress = ptr->page_table;
		sc->last = ptr;
		return (0);
	}
	return (EINVAL);
}

int nv_free_pages (nv, pAddress, page_count, agp_memory, priv_data)
	nv_state_t		*nv;
	void			**pAddress;
	unsigned int		page_count;
	unsigned int		agp_memory;
	void			**priv_data;
{
	nv_freebsd_state_t	*sc;
	unsigned long		addy;
	struct nv_alloc		*ptr;
	int			i;

	sc = nv->os_state;
	addy = (unsigned long)*pAddress;

	dprintf ("nv_free_pages\n");
	for (i = 0; i < MAX_KEYS; ++i) {
		ptr = sc->allocs[i];
		if (!ptr)
			continue;
		if (ptr->vaddr != addy)
			continue;
		kmem_free (kernel_map, ptr->vaddr, ptoa(ptr->pages));
		sc->allocs[i] = 0;
		free (ptr->page_table, M_NV);
		free (ptr, M_NV);
		return (0);
	}

	return (-1);
}

extern nv_freebsd_state_t	*lalala;

int nv_vtop (params, copy_to_user)
	nv_ioctl_memory_vtop_t	*params;
	int			copy_to_user;
{	
	void			*v;
	int			l;
	unsigned long		p, kernaddr;
	unsigned int		*physical_addresses;

	v = NV_ALIGN_PTR_DOWN(params->buffer, NVHW_PAGE_SIZE);
	l = NV_ALIGN_UP(params->byte_length + (params->buffer - v), NVHW_PAGE_SIZE);
	dprintf ("nv_vtop %d %p %x\n", copy_to_user, v, l);
	if (inc_entry())
		return (0x0);

	for (physical_addresses = params->physical_addresses; l > 0; physical_addresses++) {
		dprintf ("%p -> ", v); 
		if (v < VM_MIN_KERNEL_ADDRESS) {
			if (!curproc || !curproc->p_vmspace) {
				printf ("HMM!\n");
				return (EINVAL);
			}
			nv_find_alloc (lalala, (unsigned long)v, &kernaddr, 0);
			if (kernaddr) {
				p = vtophys (kernaddr);
			} else {
				p = pmap_extract(&(curproc->p_vmspace->vm_pmap), (vm_offset_t)v);
			}
		} else {
			p = vtophys (v);
		}
		dprintf ("%08lx\n", p);
		if (copy_to_user) {
			copyout(&p, physical_addresses, 4);
		} else {
			*physical_addresses = p;
		}

		l -= PAGE_SIZE;
		v += PAGE_SIZE;
	}

	return (0);
}

void nv_post_event (file_void)
	void			*file_void;
{
	dprintf ("nv_post_event\n");
}

void nv_post_vblank (nv)
	nv_state_t		*nv;
{
	dprintf ("nv_post_vblank\n");
}

/*
static void *
nv_malloc (unsigned int size)
{
	printf ("nv_malloc\n");
	return NULL;
	return malloc(size, M_NV, M_NOWAIT);
}

static void
nv_free (void **page_list)
{
	printf ("nv_free\n");
}
*/

/* driver module */

/*
 * The following are the supported pci device_ids
 */

#define NVIDIA_VENDOR 0x10DE

static unsigned short pci_nvidia_ids[] = {
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29,
	0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F, 0xA0, 0x100, 0x101, 0x102,
	0x103, 0x110, 0x111, 0x112, 0x113, 0x150, 0x151, 0x152, 0x153, 0
};

static int nv_probe (dev)
	device_t		dev;
{
	unsigned short		*id;
	unsigned short		device;

	dprintf("nv_probe\n");

	if (pci_get_vendor (dev) == NVIDIA_VENDOR) {
		device = pci_get_device (dev);
		for (id = pci_nvidia_ids; *id; ++id) {
			if (*id == device) {
				device_set_desc (dev, "NVIDIA");
				return (0);
			}
		}
	}

	return (ENXIO);
}

void nv_release_resources (dev)
	device_t		dev;
{
	nv_freebsd_state_t	*sc = nv_device_get_softc_os (dev);

	dprintf ("NV: nv_release_resources\n");

	if (sc->reg.res)
		bus_release_resource
			(dev, SYS_RES_MEMORY, sc->reg.rid, sc->reg.res);
	if (sc->ffb.res)
		bus_release_resource
			(dev, SYS_RES_MEMORY, sc->ffb.rid, sc->ffb.res);
	if (sc->intr.res)
		bus_release_resource
			(dev, SYS_RES_IRQ, 0, sc->intr.res);

	sc->reg.res = sc->ffb.res = sc->intr.res = NULL;
}

static int nv_attach (dev)
	device_t		dev;
{
	nv_state_t		*sc;
	nv_freebsd_state_t	*sc_os;
	u_int32_t		command;
	int			error = ENXIO;
	int			dev_num;

	/* i allocate space for both structures.. */
	sc = nv_device_get_softc (dev);
	sc->os_state = (unsigned char *)sc + sizeof(nv_state_t);
	sc_os = (nv_freebsd_state_t *)(sc->os_state);

	sc->bus = pci_get_bus (dev);
	sc->slot = pci_get_slot (dev);
	sc->vendor_id = pci_get_vendor (dev);
	sc->device_id = pci_get_device (dev);
	sc->device_number = dev_num = device_get_unit (dev);

	if (dev_num >= NV_MAX_DEVICES) {
		printf("nvidia%d: too many devices!\n", dev_num);
		goto fail;
	}

	command = pci_read_config (dev, PCIR_COMMAND, 4);
	command |= PCIM_CMD_MEMEN | PCIM_CMD_BUSMASTEREN | PCIM_CMD_PORTEN;
	pci_write_config (dev, PCIR_COMMAND, command, 4);
	command = pci_read_config (dev, PCIR_COMMAND, 4);

	if (!(command & PCIM_CMD_MEMEN)) {
		printf("nvidia%d: can't enable I/O memory!\n", dev_num);
		goto fail;
	}

	sc_os->reg.rid = 0x10;
	sc_os->ffb.rid = 0x14;
	sc_os->reg.res = sc_os->ffb.res = NULL;

	sc->reg_physical_address =
		bus_get_resource_start (dev, SYS_RES_MEMORY, sc_os->reg.rid);
	sc->fb_physical_address =
		bus_get_resource_start (dev, SYS_RES_MEMORY, sc_os->ffb.rid);

	printf ("reg_physical_address = %08lx\n", sc->reg_physical_address);
	printf ("fb_physical_address = %08lx\n", sc->fb_physical_address);

	/* only becomes usable when the device is mapped, with nv_map_device */

	sc_os->intr.rid = 0x0;
	sc_os->intr.res = NULL;

	sc->flags = 0;

	sc->interrupt_line = pci_get_irq (dev);
	sc->interrupt_pin = pci_get_intpin (dev);

	/* create the character device now */
#if 1
	sc_os->dev = make_dev(&nvidia_cdevsw, dev_num, UID_ROOT, GID_WHEEL, 0600, "nvidia%d", dev_num);
#else
	sc_os->dev=  NULL;
#endif
	error = 0;

	dprintf("dev_num: %d reg: %08lx fb: %08lx\n",
		dev_num, sc->reg_physical_address, sc->fb_physical_address);

	nv_devices[dev_num] = dev;

	sc->usage_count = 0;

	memset (sc_os->allocs, 0, sizeof(sc_os->allocs));
	sc_os->last = 0;
fail:
	return (error);
}

static int nv_detach (dev)
	device_t		dev;
{
	nv_state_t		*sc_old = nv_device_get_softc (dev);
	nv_freebsd_state_t	*sc = sc_old->os_state;
	int			num = device_get_unit (dev);
 	int			i;
	struct nv_alloc		*ptr;

	bus_generic_detach (dev);
	nv_release_resources (dev);
	if (sc->dev)
		destroy_dev (sc->dev);
	nv_devices[num] = NULL;

	for (i = 0; i < MAX_KEYS; ++i) {
		ptr = sc->allocs[i];
		if (!ptr)
			continue;
		printf ("kmem_free: %08lx\n", ptr->vaddr);
		kmem_free (kernel_map, ptr->vaddr, ptoa(ptr->pages));
		free (ptr->page_table, M_NV);
		free (ptr, M_NV);
		sc->allocs[i] = 0;
        }
	
	return (0);
}

static int nv_shutdown (dev)
	device_t	dev;
{
	/* stop all dma transfers... */
	return (0);
}

static device_method_t nv_methods[] = {
	/* Device Interface */
	DEVMETHOD(device_probe,		nv_probe),
	DEVMETHOD(device_attach,	nv_attach),
	DEVMETHOD(device_detach,	nv_detach),
	DEVMETHOD(device_shutdown,	nv_shutdown),
	{ 0, 0 }	
};

static driver_t nv_driver = {
	"NVdriver",
	nv_methods,
	/* allocate space for both structures :) */
	sizeof(nv_state_t) + sizeof(nv_freebsd_state_t) + 16
};

static devclass_t nv_devclass;

DRIVER_MODULE(NVdriver, pci, nv_driver, nv_devclass, 0, 0);
