#include "nv_ctl.h"
#include "nv-freebsd.h"
#include "nv_ref.h"
#include <machine/bus.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/module.h>
#include <sys/proc.h>
#include <sys/conf.h>
#include <sys/lock.h>
#include <sys/vnode.h>
#include <sys/sysproto.h>
#include <sys/mman.h>
#include <sys/file.h>
#include <sys/sysent.h>
#include <sys/syscall.h>
#define RM_HEADERS_MINIMAL                                                      
#include "nv.h"
#include "nvrm.h"
#include "os-interface.h"

static int nv_open		__P((dev_t, int, int, struct proc *));
static int nv_close		__P((dev_t, int, int, struct proc *));
static int nv_real_ioctl	__P((dev_t, unsigned long, unsigned long *, int, struct proc *));
static int nv_ioctl		__P((dev_t, unsigned long, caddr_t, int, struct proc *));
static int nv_mmap		__P((dev_t, vm_offset_t, int));
static int nv_poll		__P((dev_t, int, struct proc *));

static int nv_modevent		__P((module_t, int, void *));

extern device_t			nv_devices[NV_MAX_DEVICES];
nv_state_t			nv_ctl_device;
nv_freebsd_state_t		nv_fbsd_ctl_device;
unsigned long			last_ioctl_command;

static int nv_open (dev, oflags, devtype, p)
	dev_t			dev;
	int			oflags;
	int			devtype;
	struct proc		*p;
{

	dvprintf ("Control!\n");
	nv_ctl_device.device_number = NV_CONTROL_DEVICE_NUMBER;
	nv_ctl_device.flags |= NV_FLAG_OPEN;
	nv_ctl_device.flags |= NV_FLAG_CONTROL;
	nv_ctl_device.usage_count++;

	return (0);
}

static int nv_close (dev, oflags, devtype, p)
	dev_t			dev;
	int			oflags;
	int			devtype;
	struct proc		*p;
{
	dvprintf ("nv_close\n");

	if (--nv_ctl_device.usage_count)
		nv_ctl_device.flags = 0;

	nv_unix_free_all_unused_clients ((nv_state_t*)&nv_ctl_device,p->p_pid,NULL);

	return (0);
}

static int nv_real_ioctl (dev, cmd, addy, flag, p)
	dev_t			dev;
	unsigned long		cmd;
	unsigned long *		addy;
	int			flag;
	struct proc		*p;
{
	int			i;
	int			ret;
	device_t		device;
	nv_state_t		*nv;
	nv_ioctl_card_info_t	ci[NV_MAX_DEVICES];
	nv_ioctl_primary_card_t	params;

	if (inc_entry())
		return (EINVAL);

	switch (cmd) {
	case NV_IOCTL_CARD_INFO:
		dprintf ("NV_IOCTL_CARD_INFO\n");

		memset (ci, 0, sizeof(ci));

		for (i = 0; i < NV_MAX_DEVICES; ++i) {
			device = nv_devices[i];
			if (device == NULL)
				continue;
			nv = device_get_softc (device);
			ci[i].flags = NV_IOCTL_CARD_INFO_FLAG_PRESENT;; 
			ci[i].bus = nv->bus;
			ci[i].slot = nv->slot;
			ci[i].vendor_id = nv->vendor_id;
			ci[i].device_id = nv->device_id;
			ci[i].interrupt_line = nv->interrupt_line;
		}
		return (copyout (ci, addy, sizeof(ci)));

	case NV_IOCTL_PRIMARY_CARD:
		if (copyin (addy, &params, sizeof(params)) != 0)
			return (EINVAL);

		dprintf ("NV_IOCTL_PRIMARY_CARD: %d\n", params.index);
		if (params.index >= NV_MAX_DEVICES) {
			printf ("NV_IOCTL_PRIMARY_CARD, params.index is out of range\n");
			return (EINVAL);
		}

		device = nv_devices[params.index];
		if (device == NULL) {
			printf("NV_IOCTL_PRIMARY_CARD on non existing card!\n");
			return (EINVAL);
		}
		nv = device_get_softc (device);
		if (nv == NULL) {
			printf ("NV_IOCTL_PRIMARY_CARD on non existing card\n");
			return (EINVAL);
		}
		nv->flags |= NV_FLAG_POSTED;
		return (0);

	default:
		dvprintf ("nvExtEscape\n");
		if (!(nv_ctl_device.flags & NV_FLAG_CONTROL))
			dprintf ("NO CONTROL!\n");
		ret = nvExtEscape (NULL, &nv_ctl_device, cmd, addy) ? 0 : EINVAL;
		if (ret == EINVAL)
			dprintf ("nvExtEscape returned error\n");
		return (ret);
	}

	return (0);
}

static int nv_ioctl (dev, cmd, data, flag, p)
	dev_t			dev;
	unsigned long		cmd;
	caddr_t			data;
	int			flag;
	struct proc		*p;
{
	unsigned long *		addy;

	if (cmd != DUMMY_IOCTL)
		return (EINVAL);

	addy = *(unsigned long **)data;

        return (nv_real_ioctl (dev, last_ioctl_command, addy, flag, p));
}

static int nv_mmap (dev, offset, prot)
	dev_t			dev;
	vm_offset_t		offset;
	int			prot;
{
	printf ("nv_ctl_mmap\n");

	return (EINVAL);
}

static int nv_poll (dev, events, p)
	dev_t			dev;
	int			events;
	struct proc		*p;
{
	printf ("nv_ctl_poll\n");

	return (EINVAL);
}

struct cdevsw nv_cdevsw = {
	nv_open,
	nv_close,
	noread,
	nowrite,
	nv_ioctl,
	nv_poll,
	nv_mmap,
	nostrategy,
	"nvidia_ctl",
	NV_MAJOR_DEVICE_NUMBER,
	nodump,
	nopsize,
	D_TRACKCLOSE,
	-1
};

int my_mmap		__P((struct proc *, struct mmap_args *));

struct nv_alloc				*last = NULL;

void *modify_offset (udev_t udev)
{
	device_t			device;
	nv_state_t			*sc_old;
	unsigned char			key;

	device = nv_devices[uminor(udev)];
	if (!device)
		return (NULL);

	sc_old = nv_device_get_softc (device);
	if (!sc_old)
		return (NULL);

	key = nv_alloc_get_key (sc_old);
	if (key == -1)
		return (NULL);

	return ((void *)NV_ALLOC_FROM_KEY(key));
}

int my_ioctl (p, uap)
	struct proc			*p;
	register struct ioctl_args	*uap;
{
	register struct filedesc	*fdp = p->p_fd;
	register struct file		*fp = NULL;
	struct vnode			*vp;
	int				ret;

	if (((unsigned)uap->fd) >= fdp->fd_nfiles || (fp = fdp->fd_ofiles[uap->fd]) == NULL) {
		goto out;
	}

	if (fp->f_type != DTYPE_VNODE)
		goto out;

	vp = (struct vnode *)fp->f_data;
	if (vp->v_type != VCHR)
		goto out;

	if (umajor(vp->v_rdev->si_udev) != 195)
		goto out;

	last_ioctl_command = uap->com;	
	uap->com = DUMMY_IOCTL;

out:
	ret = ioctl (p, uap);
	return (ret);
}

int my_mmap (p, uap)
	struct proc 			*p;
	register struct mmap_args	*uap;
{
	register struct filedesc 	*fdp = p->p_fd;
	register struct file 		*fp = NULL;
	struct vnode 			*vp;
	int				flag = 0, ret;

	if (!(uap->flags & MAP_ANON)) {
                if (((unsigned) uap->fd) >= fdp->fd_nfiles ||
                    (fp = fdp->fd_ofiles[uap->fd]) == NULL) {
			goto out;
		}
		if (fp->f_type != DTYPE_VNODE)
			goto out;

		vp = (struct vnode *) fp->f_data;
		if (vp->v_type != VCHR)
			goto out;
		if (umajor(vp->v_rdev->si_udev) == 195) {
			dprintf ("mmap address: %p len: %x offset: %x\n", uap->addr, uap->len, uap->pos);
			flag++;
			if (uap->pos == NV_MMAP_ALLOCATION_OFFSET) {
				uap->pos = modify_offset (vp->v_rdev->si_udev);
				if (uap->pos == NULL)
					return (EINVAL);
			}
			if (inc_entry())
				return (EINVAL);
		}
	}
out:
	ret = mmap (p, uap);
	if (flag && !ret) {
		dvprintf ("return: %08x\n", p->p_retval[0]);
		if (last)
			last->uvaddr = p->p_retval[0];
	}
	return (ret);
}

static int nv_modevent (mod, type, data)
	module_t		mod;
	int			type;
	void			*data;
{
	static dev_t		ctl_dev = 0;

	switch (type) {
	case MOD_LOAD:
		if (!RmInitRm()) {
			printf ("RmInitRm failed\n");
			return (EINVAL);
		}
		nv_ctl_device.flags = 0;
		nv_ctl_device.usage_count = 0;
		nv_ctl_device.os_state = (void *)&nv_fbsd_ctl_device;
		
		ctl_dev = make_dev (&nv_cdevsw, 255, UID_ROOT, GID_WHEEL, 0600, "nv_ctl");
		sysent[SYS_mmap].sy_call = (sy_call_t *)my_mmap;
		sysent[SYS_ioctl].sy_call = (sy_call_t *)my_ioctl;
		break;
	case MOD_UNLOAD:
		RmShutdownRm();
		if (ctl_dev)
			destroy_dev (ctl_dev);
		sysent[SYS_mmap].sy_call = (sy_call_t *)mmap;
		sysent[SYS_ioctl].sy_call = (sy_call_t *)ioctl;
		break;
	}
	return (0);
}

DEV_MODULE(nv, nv_modevent, NULL);
